/*
 * Copyright  2005 Vyacheslav Anikin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: rux.c,v 1.24 2005/11/25 20:48:03 ghos Exp $ */

/* VERSION was defined in Makefile */
static char *version =
#ifndef VERSION
"float!";

#else
VERSION;

#endif					/* not VERSION */

#include <sys/types.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <libgen.h>

#include "cptaba.h"
#include "cpdetect.h"

u_char *init_tab(u_char * tab, int len, struct cpinfo *in, struct cpinfo *out);
struct cpinfo *get_cp_info(const char *cpid);
void print_usage(void);
void print_cplist(struct cpinfo *cplist);

static int gr = 0;

#define DEFANSIZE 10240

static struct cpinfo *i_cp = NULL;
static struct cpinfo *o_cp = NULL;

int main(int argc, char **argv)
{
	u_char conv_tab[256];

	int c;				/* input/output buffer */
	int fnoconv = 0;		/* don't convert by default */
	int fdetect = 1;		/* auto detect by default */

	int ansize = DEFANSIZE;		/* buffer size for analyzing */
	const int max_ansize = 102400;	/* maximum buffer size */

	FILE *in = stdin;
	FILE *out = stdout;

	extern char *optarg;
	extern int optind;

	o_cp = &codepage[2];		/* sets koi8-r as default outcp */
	int flag;

	while ((flag = getopt(argc, argv, "i:o:thegTs:")) != -1) {
		switch (flag) {
		case 'e':
			gr = 0;
			break;
		case 'g':
			gr = 1;
			break;
		case 'i':
			i_cp = get_cp_info(optarg);
			fdetect = 0;
			break;
		case 'o':
			o_cp = get_cp_info(optarg);
			break;
		case 't':
			fnoconv = 0;
			fdetect = 1;
			break;
		case 'T':
			fnoconv = 1;
			fdetect = 1;
			break;
		case 's':
			if ((ansize = strtol(optarg, NULL, 10)) > max_ansize) {
				ansize = (ansize <= 0) ? DEFANSIZE : max_ansize;
			}
			break;
		case 'h':
		default:
			print_usage();
			return 0;
		}
	}

	argc -= optind;
	argv += optind;

	if (o_cp == NULL) {
		printf("rux: error: missing or corrupt output code page id!\n");
		print_cplist(codepage);
		return 2;
	}
	if (fdetect == 0) {
		if (i_cp == NULL) {
			printf("rux: error: missing or corrupt input code page id!\n");
			print_cplist(codepage);
			return 3;
		}
	}
	char *buf = (char *)malloc(ansize);

	if (buf == NULL) {
		perror(NULL);
		return errno;
	}
	int readstdin = 0;		/* read from stdin */

	if (argc == 0)
		readstdin = 1;

	while (1) {
		if (readstdin) {
			in = stdin;
		} else if ((in = fopen(*argv, "r")) == NULL) {
			perror(*argv);
			return errno;
		}
		int buflen = 0;

		if (fdetect) {
			buflen = fread(buf, 1, ansize, in);

			/* The detect_cp always returns a valid code page id */
			/* Default output code page is user selected o_cp */
			i_cp = detect_cp(codepage, o_cp, buf, buflen);
			if (fnoconv) {
				printf("%s: %s\n", readstdin ? "<stdin>" : basename(*argv),
					i_cp->cp_name);
			}
		}
		if (fnoconv == 0) {
			init_tab(conv_tab, sizeof(conv_tab), i_cp, o_cp);
			if (fdetect) {
				u_char *p = (u_char *) buf;

				while (buflen--) {
					putc(conv_tab[*p++], out);
				}
			}
			while ((c = getc(in)) != EOF)
				putc(conv_tab[(u_char) c], out);
		}
		fclose(in);
		if (readstdin)
			break;

		if (*++argv == NULL)
			break;
	}

	free(buf);
	return 0;
}

struct cpinfo *get_cp_info(const char *cpid)
{
	int i = 0;

	while (1) {
		if (codepage[i].cp_name == NULL)
			break;
		if (strcmp(codepage[i].cp_name, cpid) == 0)
			return &codepage[i];
		++i;
	}
	return NULL;
}

void compl_tab(u_char * tab, struct cpinfo *incp, struct cpinfo *outcp)
{
	int i;

	if ((i = incp->cp_graphsize) != 0) {
		for (--i; i >= 0; i--) {
			tab[incp->cp_gdata[i]] = gr ? outcp->cp_gdata[i] : boxdr_map[i];
		}
	}
}

u_char *init_tab(u_char * tab, int len, struct cpinfo *in, struct cpinfo *out)
{
	int i = len - 1;

	for (; i >= 0; i--) {
		tab[i] = i;
	}

	for (i = out->cp_size; i >= 0; i--) {
		tab[in->cp_data[i]] = out->cp_data[i];
	}

	compl_tab(tab, in, out);

	return tab;
}

void print_usage(void)
{
	printf(" rux-%s " DATETIME "\n"
		" usage: rux [-egtT] [-s size] [-i incp] [-o outcp] [file ...]\n\n"
		"    -h     -- this help message\n"
		"    -i     -- input code page id\n"
		"    -o     -- output code page id [koi8-r]\n"
		"    -e     -- use universal translating of the"
		" box-drawing chars [default]\n"
		"    -g     -- inverse of -e\n"
		"    -t     -- detect an input code page [default]\n"
		"    -T     -- just detect an input code page and exit\n"
		"    -s     -- buffer size in bytes for analyzing [10240]\n\n",
		version
	);
	print_cplist(codepage);
	printf("\nVyachelav Anikin <ghos@mail.ru>\n");
}

void print_cplist(struct cpinfo *cplist)
{
	struct cpinfo *curcp = cplist;

	printf("The valid code pages identifiers are:");

	for (; curcp->cp_name; curcp++) {
		printf(" %s", curcp->cp_name);
	}

	printf("\n");
}
